/** @file
 * Low-level function to receive and parse a reply from an ident server.
 *
 * @author Peter Eriksson <pen@lysator.liu.se> with modifications by Pr
 *   Emanuelsson <pell@lysator.liu.se>.  Support for S/Ident protocol added by
 *   Booker Bense <bbense@stanford.edu>.
 *
 * $Id: id_parse.c,v 1.4 2002/10/29 00:51:20 eagle Exp $
 */

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#ifdef HAVE_SYS_SELECT_H
# include <sys/select.h>
#endif

#include "sident.h"

extern char *id_strtok();

/** Returns the string corresponding to an error response code. */
int
ident_result_code(char *error_string)
{
  int i, res ; 
  const char *error_text ; 
  res = IDENT_UNKNOWN_ERROR ; 
  for( i = 0 ; i < IDENT_MAX_ERROR ; i++ ) {
    error_text = ident_err_txt[i] ; 
    if (strncmp(error_string,error_text,strlen(error_text)) == 0 ) { 
      return i ;
    } 

  } 
  
 return res ; 


}

/** Read a line of data from the network with an optional timeout.
 *
 * The line of data will be left in id->buf on success.
 *
 * @param id The open connection to the responder.
 * @param timeout Timeout if desired, interpreted the same as select(2).
 * @return The same codes as would be returned by id_parse.
 */
int
id_read_response(ident_t *id, struct timeval *timeout)
{
    fd_set read_set;
    int status;
    size_t offset;

    if (id == NULL)
        return IDENT_SYSTEM_ERROR;

    /* If timeout is set, use select to be sure there's data there to read. */
    if (timeout != NULL) {
	FD_ZERO(&read_set);
	FD_SET(id->fd, &read_set);

        status = select(FD_SETSIZE, &read_set, NULL, NULL, timeout);
        if (status < 0)
            return IDENT_SYSTEM_ERROR;
        if (status == 0) {
            errno = ETIMEDOUT;
            return IDENT_TIMEOUT;
        }
    }

    /* This is completely broken; it doesn't actually implement timeouts. */
    offset = 0;
    while (offset < sizeof(id->buf)) {
        status = read(id->fd, id->buf + offset, sizeof(id->buf) - offset - 1);
        offset += status;
        id->buf[offset] = '\0';
#ifdef DEBUG
        printf("Read response %ld %s\n", offset, id->buf);
#endif
        if (status < 0)
            return IDENT_SYSTEM_ERROR;
        if (id->buf[offset - 1] == '\n' || id->buf[offset - 1] == '\r')
            break;
        if (status == 0) {
            errno = ENOTCONN;
            return IDENT_SYSTEM_ERROR;
        }
    }
    return IDENT_AUTH_OKAY;
}

int
id_parse(ident_t * id, struct timeval *timeout, int *resp_port,
         int *req_port, char **identifier, char **opsys, char **charset,
         struct ident_auth_client_data *auth_data)
{
    char c, *cp,*tmp_cp,*tmp_charset;
    fd_set rs;
    int pos, res=0, resp, reqp, rcode ;
    
    errno = 0;
    
    tmp_charset = 0;
    
    if (!id)
	return -1;
    if (resp_port)
	*resp_port = 0;
    if (req_port)
	*req_port = 0;
    
    if ( identifier != NULL ) 
      *identifier = NULL;
    if ( opsys != NULL ) 
      *opsys = NULL ;
    if ( charset != NULL ) 
      *charset = NULL ;

    res = id_read_response(id, timeout);
    if (res != IDENT_AUTH_OKAY)
        return res;
    
    /*
    ** Get first field (<resp_port> , <req_port>)
    */
    cp = id_strtok(id->buf, ":", &c);
    if (!cp)
	return IDENT_INVALID_RESP_INFO;
    
    if (sscanf(cp, " %d , %d", &resp, &reqp) != 2)
    {
	if (identifier)
	{
	    *identifier = strdup(cp);
	    if (*identifier == NULL)
	        return IDENT_SYSTEM_ERROR;
	}
	return  IDENT_INVALID_RESP_INFO;
    }
    
    if (resp_port)
	*resp_port = resp;
    if (req_port)
	*req_port = reqp;

    /*
    ** Get second field (USERID , ERROR , AUTHENTICATE)
    */
    cp = id_strtok((char *)0, ":", &c);
    if (!cp)
	return IDENT_INVALID_RESP_INFO;
    
    if (strcmp(cp, "ERROR") == 0)
    {
	cp = id_strtok((char *)0, "\n\r", &c);
	if (!cp)
	    return IDENT_INVALID_RESP_INFO;
	
	if (identifier)
	{
	    *identifier = strdup(cp);
	    if (*identifier == NULL)
	        return  IDENT_SYSTEM_ERROR;
	}
	
	return(ident_result_code(*identifier));
    }
    else if (strcmp(cp, "USERID") == 0)
    {
	/*
	** Get first subfield of third field <opsys>
	*/
	cp = id_strtok((char *) 0, ",:", &c);
	if (!cp)
	    return IDENT_INVALID_RESP_INFO;
	if ( opsys != NULL ) {
	  *opsys = strdup(cp);
	  if (*opsys == NULL)
	    return IDENT_SYSTEM_ERROR;
	}
	
	/*
	** We have a second subfield (<charset>)
	*/
	if (c == ',')
	{
	    cp = id_strtok((char *)0, ":", &c);
	    if (!cp)
		return IDENT_INVALID_RESP_INFO;
	    
	    tmp_charset = cp;
	    if (charset)
	    {
		*charset = strdup(cp);
		if (*charset == NULL)
		    return IDENT_SYSTEM_ERROR;
	    }
	    
	    /*
	    ** We have even more subfields - ignore them
	    */
	    if (c == ',')
		id_strtok((char *)0, ":", &c);
	}
	
	if (tmp_charset && strcmp(tmp_charset, "OCTET") == 0)
	    cp = id_strtok((char *)0, (char *)0, &c);
	else
	    cp = id_strtok((char *)0, "\n\r", &c);
	
	if (identifier != NULL )
	{
	    *identifier = strdup(cp);
	    if (*identifier == NULL)
	        return IDENT_SYSTEM_ERROR;
	}
	return IDENT_AUTH_NOT_SUPPORTED ;
    } else if (strcmp(cp, "AUTHENTICATE") == 0) { 
      tmp_cp = id_strtok((char *)0, "\n\r", &c) ; 
      strncpy(auth_data->authenticate_data,tmp_cp,MAX_AUTH_LEN);
      rcode = (*auth_data->auth_struct->auth)(auth_data);
      if (rcode == IDENT_AUTH_OKAY ) { 
	*identifier = strdup(auth_data->user_ident) ; 
      } else { 
        *identifier = strdup(cp) ; 
      }
      return rcode ; 
    } 
    else
    {
	if (identifier)
	{
	    *identifier = strdup(cp);
	    if (*identifier == NULL)
	        return IDENT_SYSTEM_ERROR;
	}
	return IDENT_UNKNOWN_ERROR;
    }
}
