/** @file
 * A TCP/IP S/Ident protocol server.
 *
 * This program is in the public domain and may be used freely by anyone who
 * wants to.
 *
 * @date 1993-10-07
 * @author Peter Eriksson <pen@lysator.liu.se>
 * @author extensively modified for S/Ident by Booker C. Bense
 *   <bbense@stanford.edu>.
 *
 * $Id: identd.c,v 1.4 2003/05/30 23:45:16 eagle Exp $
 */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <netdb.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

/* BSDI needs <netinet/in.h> before <arpa/inet.h>. */
#include <netinet/in.h>
#include <arpa/inet.h>

#if HAVE_SYS_SELECT_H
# include <sys/select.h>
#endif

#include "error.h"
#include "identd.h"
#include "paths.h"

char *path_unix = NULL;
char *path_kmem = NULL;

int verbose_flag = 0;
int debug_flag   = 0;
int syslog_flag  = 0;
int multi_flag   = 0;
int other_flag   = 0;
int unknown_flag = 0;
int uidonly_flag = 0 ; 
int users_can_nosident = ( 0 == 0 ) ;
int allow_trivial = 0 ; 
int mutual_auth_flag = 0 ; 
int client_only_flag = 0 ; 

int lport = 0;
int fport = 0;

char *charset_name = NULL;

static int child_pid;

static int syslog_facility = LOG_DAEMON;

/*
** The structure passing convention for GCC is incompatible with
** Suns own C compiler, so we define our own inet_ntoa() function.
** (This should only affect GCC version 1 I think, a well, this works
** for version 2 also so why bother.. :-)
*/
#if defined(__GNUC__) && defined(__sparc__) && !defined(NeXT)

#ifdef inet_ntoa
#undef inet_ntoa
#endif

char *inet_ntoa(ad)
    struct in_addr ad;
{
    unsigned long int s_ad;
    int a, b, c, d;
    static char addr[20];
    
    s_ad = ad.s_addr;
    d = s_ad % 256;
    s_ad /= 256;
    c = s_ad % 256;
    s_ad /= 256;
    b = s_ad % 256;
    a = s_ad / 256;
    sprintf(addr, "%d.%d.%d.%d", a, b, c, d);
    
    return addr;
}
#endif

static int comparemem(p1, p2, len)
    unsigned char *p1;
    unsigned char *p2;
    int len;
{
    int c;

    while (len-- > 0)
	if ((c = (int) *p1++ - (int) *p2++) != 0)
	    return c;

    return 0;
}

/*
** Return the name of the connecting host, or the IP number as a string.
*/
char *gethost(addr)
    struct in_addr *addr;
{
    int i;
    struct hostent *hp;
    char *hostname;
  
    hp = gethostbyaddr((char *) addr, sizeof(struct in_addr), AF_INET);
    if (hp)
    {
	/* Found a IP -> Name match, now try the reverse for security reasons */
        hostname = strdup(hp->h_name);
        if (hostname == NULL)
            return inet_ntoa(*addr);
	hp = gethostbyname(hostname);
        free(hostname);
	if (hp)
	    for (i = 0; hp->h_addr_list[i]; i++)
		if (comparemem(hp->h_addr_list[i],
			       (unsigned char *) addr,
			       sizeof(struct in_addr)) == 0)
		    return hp->h_name;
  }

  return inet_ntoa(*addr);
}


char *clearmem(bp, len)
    char *bp;
    int len;
{
    char *cp;

    cp = bp;
    while (len-- > 0)
	*cp++ = 0;
    
    return bp;
}


/*
** Main entry point into this daemon
*/
int main(argc,argv)
    int argc;
    char *argv[];
{
    int i, len;
    struct sockaddr_in sin;
    struct in_addr laddr, faddr;
    struct timeval tv;
    int one = 1;

    int background_flag = 0;
    int timeout = 0;
    char *portno = "113";
    char *bind_address = NULL;
    int set_uid = 0;
    int set_gid = 0;
    int inhibit_default_config = 0;
    int opt_count = 0;		/* Count of option flags */
  
  
    /*
    ** Parse the command line arguments
    */
    for (i = 1; i < argc && argv[i][0] == '-'; i++) {
	opt_count++;
	switch (argv[i][1])
	{

	  case 'i':    /* Start from Inetd, nowait mode */
	    background_flag = 0;
	    break;
	    
	  case 't':
	    timeout = atoi(argv[i]+2);
	    break;
	    
	  case 'p':
	    portno = argv[i]+2;
	    break;
	    
	  case 'a':
	    bind_address = argv[i]+2;
	    break;
	    
	  case 'c':
	    charset_name = argv[i]+2;
	    break;
	    
	    
	  case 'l':    /* Use the Syslog daemon for logging */
	    syslog_flag++;
	    break;
	      
	case 'n':
            uidonly_flag = 1;
            break;
            

	    
	  case 'o':
	    other_flag = 1;
	    break;
	    
	  case 'e':
	    unknown_flag = 1;
	    break;
	    
	  case 'V':    /* Give version of this daemon */
	    printf("[%s]\r\n", PACKAGE_STRING);
	    exit(0);
	    break;
	    
	  case 'v':    /* Be verbose */
	    verbose_flag++;
	    break;
	    
	  case 'd':    /* Enable debugging */
	    debug_flag++;
	    break;
	
	  case 'N':    /* Disable users ".nosident" files */
	    users_can_nosident = ( 0 == 1 ) ;
	    break;
  	  case 'M':      /* Turn off mutual auth */ 
	    mutual_auth_flag++ ; 
	    break ; 

	  case 'T':       /* Allow trivial auth */ 
	    allow_trivial++ ; 
	    break;
	  case 'C':         /* Turn on client only port checking */ 
	    client_only_flag++ ; 
	    break; 
 
	  default:
	    ERROR1("Bad option %s", argv[i]);
	    break;
	}
    }

    /*
    ** Path to kernel namelist file specified on command line
    */
    if (i < argc)
	path_unix = argv[i++];

    /*
    ** Path to kernel memory device specified on command line
    */
    if (i < argc)
	path_kmem = argv[i++];


    if (i < argc)
	ERROR1("Too many arguments: ignored from %s", argv[i]);


    /*
    ** We used to call k_open here. But then the file descriptor
    ** kd->fd open on /dev/kmem is shared by all child processes.
    ** From the fork(2) man page:
    **
    ** o The child process has its own copy of the parent's descriptors.
    **   These descriptors reference the same underlying objects.  For
    **   instance, file pointers in file objects are shared between the
    **   child and the parent so that an lseek(2) on a descriptor in the
    **   child process can affect a subsequent read(2) or write(2) by
    **   the parent.
    **
    ** Thus with concurrent (simultaneous) identd client processes,
    ** they step on each other's toes when they use kvm_read.
    ** 
    ** Calling k_open here was a mistake for another reason too: we
    ** did not yet honor -u and -g options. Presumably we are
    ** running as root (unless the in.identd file is setuid), and
    ** then we can open kmem regardless of -u and -g values.
    ** 
    ** 
    ** Open the kernel memory device and read the nlist table
    ** 
    **     if (k_open() < 0)
    ** 		ERROR("main: k_open");
    */

    /*
    ** Get foreign internet address
    */
    len = sizeof(sin);
    if (getpeername(0, (struct sockaddr *) &sin, &len) == -1)
    {
	/*
	** A user has tried to start us from the command line or
	** the network link died, in which case this message won't
	** reach to other end anyway, so lets give the poor user some
	** errors.
	*/
	perror("in.sidentd: getpeername()");
	exit(1);
    }
    
    faddr = sin.sin_addr;
    
    
    /*
    ** Open the connection to the Syslog daemon if requested
    */
    if (syslog_flag)
    {
	openlog("sidentd", LOG_PID, syslog_facility);
    }
  

    /*
    ** Get local internet address
    */
    len = sizeof(sin);
    if (getsockname(0, (struct sockaddr *) &sin, &len) == -1)
    {
	/*
	** We can just die here, because if this fails then the
	** network has died and we haven't got anyone to return
	** errors to.
	*/
	exit(1);
    }
    laddr = sin.sin_addr;
    

    /*
    ** Get the local/foreign port pair from the luser
    */
    parse(stdin, &laddr, &faddr);

    /* Note : we should never return from parse */ 
    exit(0);
}
